"""
This file contains all the functions that could be commonly used to convert
dates to strings.  This is a very important and necessary tool in climate
observation and reanalysis data, because files are often saved in certain
formats that include the date and time of the observation.

There following types of functions are contained in this .jl module:
1)  single-unit conversion functions (e.g. yr2str) that extracts only a single
    time-unit and converts it to a string
2)  multi-unit conversion functions (e.g. ymd2str) that extracts and combines
    various different time units into a string
3)  conversion of dates into folder directories

"""

# Single-Unit String Functions
function yr2str(dnt::TimeType)
    return @sprintf("%04d",Dates.year(dnt));
end

function mo2str(dnt::TimeType)
    return @sprintf("%02d",Dates.month(dnt));
end

function dy2str(dnt::TimeType)
    return @sprintf("%02d",Dates.day(dnt));
end

function hr2str(dnt::TimeType)
    return @sprintf("%02d",Dates.hour(dnt));
end

function mi2str(dnt::TimeType)
    return @sprintf("%02d",Dates.minute(dnt));
end

# Multi-Unit String Functions
function yrmo2str(dnt::TimeType)
    yr = Dates.year(dnt); mo = Dates.month(dnt);
    return @sprintf("%04d%02d",yr,mo);
end

function ymd2str(dnt::TimeType)
    yr = Dates.year(dnt); mo = Dates.month(dnt); dy = Dates.day(dnt);
    return @sprintf("%04d%02d%02d",yr,mo,dy);
end

function ymdhm2str(dnt::TimeType)
    yr = Dates.year(dnt); mo = Dates.month(dnt); dy = Dates.day(dnt);
    hr = Dates.hour(dnt); mi = Dates.minute(dnt);
    return @sprintf("%04d%02d%02d%02d%02d00",yr,mo,dy,hr,mi);
end

# Directory String Functions
function yrmo2dir(dnt::TimeType)
    yr = Dates.year(dnt); mo = Dates.month(dnt);
    return @sprintf("%04d/%02d",yr,mo);
end

function ymd2dir(dnt::TimeType)
    yr = Dates.year(dnt); mo = Dates.month(dnt); dy = Dates.day(dnt);
    return @sprintf("%04d/%02d/%02d",yr,mo,dy);
end

function yrdy2dir(dnt::TimeType)
    yr = Dates.year(dnt); dy = Dates.dayofyear(dnt);
    return @sprintf("%04d/%02d",yr,dy);
end

# Convert Season to Months

function dntsea2mon(season::Int)

    if     season == 0; return ':';
    elseif season == 1; return [12,1,2]; elseif season == 2; return [3,4,5];
    elseif season == 3; return [6,7,8];  elseif season == 4; return [9,10,11];
    end

end

function dntsea2mon(season::AbstractString)

    if     season == "all"; return ':';
    elseif season == "DJF"; return [12,1,2]; elseif season == "MAM"; return [3,4,5];
    elseif season == "JJA"; return [6,7,8];  elseif season == "SON"; return [9,10,11];
    end

end
